package dovecot

import (
	"context"
	"encoding/base64"
	"errors"
	"strings"

	"git.autistici.org/ai3/go-common/clientutil"

	"git.autistici.org/id/keystore/backend"
	"git.autistici.org/id/keystore/client"
	"git.autistici.org/id/keystore/userenckey"
)

// Config for the dovecot-keystore daemon.
type Config struct {
	Shard      string                    `yaml:"shard"`
	LDAPConfig *backend.LDAPConfig       `yaml:"ldap"`
	Keystore   *clientutil.BackendConfig `yaml:"keystore"`
}

// Database represents the interface to the underlying backend for
// encrypted user keys.
type Database interface {
	GetPublicKey(context.Context, string) []byte
	GetPrivateKeys(context.Context, string) [][]byte
}

func (c *Config) check() error {
	if c.LDAPConfig == nil {
		return errors.New("missing backend config")
	}
	return c.LDAPConfig.Valid()
}

type userdbResponse struct {
	PublicKey string `json:"mail_crypt_global_public_key"`
}

type passdbResponse struct {
	PrivateKey string `json:"mail_crypt_global_private_key"`
}

var passwordSep = "/"

// KeyLookupProxy interfaces Dovecot with the user encryption key database.
type KeyLookupProxy struct {
	config   *Config
	keystore client.Client
	db       Database
}

func NewKeyLookupProxy(config *Config) (*KeyLookupProxy, error) {
	if err := config.check(); err != nil {
		return nil, err
	}

	ksc, err := client.New(config.Keystore)
	if err != nil {
		return nil, err
	}

	// There is only one supported backend type, ldap.
	ldap, err := backend.NewLDAPBackend(config.LDAPConfig)
	if err != nil {
		return nil, err
	}

	return &KeyLookupProxy{
		config:   config,
		keystore: ksc,
		db:       ldap,
	}, nil
}

// Lookup a key using the dovecot dict proxy interface.
//
// We can be sent a userdb lookup, or a passdb lookup, and we can tell
// them apart from the structure of the key:
//
// If it contains passwordSep, then it's a passdb lookup and the key
// consists of 'username' and 'password'. Otherwise, it's a userdb
// lookup and the key is simply the username.
func (s *KeyLookupProxy) Lookup(ctx context.Context, key string) (interface{}, bool) {
	if strings.Contains(key, passwordSep) {
		kparts := strings.SplitN(key, passwordSep, 2)
		return s.lookupPassdb(ctx, kparts[0], kparts[1])
	}
	return s.lookupUserdb(ctx, key)
}

func (s *KeyLookupProxy) lookupUserdb(ctx context.Context, username string) (interface{}, bool) {
	pub := s.db.GetPublicKey(ctx, username)
	if pub == nil {
		return nil, false
	}
	return &userdbResponse{PublicKey: b64encode(pub)}, true
}

func (s *KeyLookupProxy) lookupPassdb(ctx context.Context, username, password string) (interface{}, bool) {
	// If the password is a SSO token, try to fetch the
	// unencrypted key from the keystore daemon.
	priv, err := s.keystore.Get(ctx, s.config.Shard, username, password)
	if err == nil {
		return &passdbResponse{PrivateKey: b64encode(priv)}, true
	}

	// Otherwise, fetch encrypted keys from the db and attempt to
	// decrypt them.
	encKeys := s.db.GetPrivateKeys(ctx, username)
	if len(encKeys) == 0 {
		return nil, false
	}
	priv, err = userenckey.Decrypt(encKeys, []byte(password))
	if err != nil {
		return nil, false
	}
	return &passdbResponse{PrivateKey: b64encode(priv)}, true
}

func b64encode(b []byte) string {
	return base64.StdEncoding.EncodeToString(b)
}
